@@ -3,23 +3,34 @@ package Tie::RangeHash;
 require 5.005;
 
 use strict;
-use warnings;
+
+use warnings::register 'Tie::RangeHash';
 
 require Carp;
 require Tie::Hash;
 
 use vars qw($VERSION $SEPARATOR @ISA);
 
-$VERSION = '0.30';
+$VERSION = '0.40';
 
 @ISA = qw(Tie::Hash);
 
-sub TYPE_NUMBER { return 1; }           # Some internal constants
-sub TYPE_STRING { return 2; }
+sub TYPE_STRING { 1; }                  # Some constants for Tie::RangeHash
+sub TYPE_NUMBER { 2; }
 
-# Default separator character ... Ceveat: use of this outside of the
-# module is deprecated! Define the SEPARATOR when using tie instead.
-$SEPARATOR; BEGIN { *SEPARATOR = \ ","; }
+BEGIN
+  {
+    # Default separator character ... Ceveat: use of this outside of the
+    # module is deprecated! Define the SEPARATOR when using tie instead.
+
+    *SEPARATOR = \ ",";
+  }
+
+sub import
+  {
+    my $class = shift;
+    
+  }
 
 sub _join_bounds                        # generate a key from bounds
   {
@@ -39,13 +50,16 @@ sub _split_bounds                       # generate bounds from a key
 
 	if (defined($extra_stuff))
 	  {
-	    Carp::carp "Multiple separators in \`$key\' will be ignored";
+	    warnings::warn
+	      "Multiple separators in \`$key\' will be ignored",
+	      if (warnings::enabled);
 	  }
 
 	if (&{$self->{COMPARE}}($lower_bound, $upper_bound) > 0)
 	  # make sure $lower_bound < $upper_bound
 	  {
-	    Carp::carp "Lower and upper bounds reversed";
+	    warnings::warn "Lower and upper bounds reversed",
+	      if (warnings::enabled);
 	    ($lower_bound, $upper_bound) = ($upper_bound, $lower_bound);
 	  }
       }
@@ -61,6 +75,11 @@ sub _new_node                           # Create a new node
   {
     my ($self, $lower_bound, $upper_bound, $value) = @_;
 
+    unless ( (defined($lower_bound)) and (defined($upper_bound)))
+      {
+	Carp::croak "Cannot create a node without valid keys";
+      }
+
     return {
       KEY_LOW    => $lower_bound,       # lower bound
       KEY_HIGH   => $upper_bound,       # upper bound
@@ -93,6 +112,10 @@ sub _add_node                           # Recursively add a new node to tree
   {
     my ($self, $root, $node) = @_;
 
+    unless ($node)
+      {
+	Carp::croak "Cannot add a NULL node";
+      }
     unless ($root) { return $node; }
 
     if ( &{$self->{COMPARE}}($node->{KEY_HIGH},  $root->{KEY_LOW} ) < 0)
@@ -107,12 +130,14 @@ sub _add_node                           # Recursively add a new node to tree
       }
     else
       {
-	#Hmm... should it carp or croak? Smart thing to do: if the nodes have
-	# different values, croak, otherwise just carp.
-	Carp::carp "Overlapping key range: cannot add range '`" .
-	  $self->_join_bounds( $node->{KEY_LOW}, $node->{KEY_HIGH}) .
-	  "\' because there exists \`" .
-	  $self->_join_bounds( $root->{KEY_LOW}, $root->{KEY_HIGH}) . "\'";
+	# Hmmm... should we warn or die here?
+	warnings::warn
+	  ("Overlapping key range: cannot add range \`" .
+	   $self->_join_bounds($node->{KEY_LOW}, $node->{KEY_HIGH}) .
+	   "\' because there exists \`" .
+	   $self->_join_bounds($root->{KEY_LOW}, $root->{KEY_HIGH}) . "\'"),
+	     if (warnings::enabled);
+
       }
 
 #    unless ($self->{UNBALANCED}) # disable tree balancing...
@@ -164,20 +189,26 @@ sub _add_new_node                       # Add a new node from the root
   }
 
 
-sub _find_node                          # Recursively find a node
+
+sub _find_node_parent                   # recursively find a node & parent
   {
-    my ($self, $root, $lower_bound, $upper_bound) = @_;
+
+    # We now also find the node and the parent node; the optimization hit
+    # is minimal compared to the advantage of having one routine that can
+    # return more info and be used by the DELETE method
+
+    my ($self, $root, $parent, $lower_bound, $upper_bound) = @_;
 
     unless ($root) { return; }
 
     if (&{$self->{COMPARE}}($upper_bound, $root->{KEY_LOW})<0)
       {
-	return _find_node($self, $root->{NODE_LEFT},
+	return _find_node_parent($self, $root->{NODE_LEFT}, $root,
 			  $lower_bound, $upper_bound);
       }
     elsif (&{$self->{COMPARE}}($lower_bound, $root->{KEY_HIGH})>0)
       {
-	return _find_node($self, $root->{NODE_RIGHT},
+	return _find_node_parent($self, $root->{NODE_RIGHT}, $root,
 			  $lower_bound, $upper_bound);
       }
     else
@@ -193,14 +224,26 @@ sub _find_node                          # Recursively find a node
 	       or (&{$self->{COMPARE}}($upper_bound, $root->{KEY_HIGH}) > 0)
 	      ) )
 	  {
-	    Carp::carp( "Key range \`", 
-		  _join_bounds($self, $lower_bound, $upper_bound),
-		  "\' exceeds defined key range \`",
-		  _join_bounds($self, $root->{KEY_LOW}, $root->{KEY_HIGH}),
-		  "\'" );
+	    warnings::warn
+	      ("Key range \`" . 
+	       _join_bounds($self, $lower_bound, $upper_bound) .
+	       "\' exceeds defined key range \`" .
+	       _join_bounds($self, $root->{KEY_LOW}, $root->{KEY_HIGH}) .
+	       "\'" ),
+		 if (warnings::enabled);
 	    return;
 	  }
-	return $root->{VALUE};
+
+	# _find_node_parent() now returns the node as opposed to the value;
+	# aside from making the function's name more accurate as to what it
+	# does, it also allows us to properly handle the case where:
+	#
+	#   $hash{'low,high'} = undef
+	#
+	# Previously, EXISTS would get the value and say the key did not
+	# exist.
+
+	return ($root, $parent);
       }
   }
 
@@ -232,16 +275,17 @@ sub TIEHASH                             # Tie::RangeHash constructor
 	$self->{SEPARATOR} = $SEPARATOR; # so we have at least something! 
       }
     else
-      {	
+      {
 	Carp::croak
-	  "\`Separator\' attribute must be a Regexp";
+	  "\`Separator\' attribute must be a SCALAR or Regexp";
       }
 
     if (defined($self->{COMPARE}))
       {
 	if (ref($self->{COMPARE}) ne "CODE")
 	  {
-	    Carp::croak "\`Comparison\' must be a CODE reference";
+	    Carp::croak
+	      "\`Comparison\' must be a CODE reference";
 	  }
       }
     else
@@ -278,18 +322,23 @@ sub FETCH                               # Retrieve a node from tree
     # number.
     # 
     # So we've modified _find_node() to handle lower and upper bounds
-    # (it will use the lower bound as the search key, and then compare
-    # the upper bound when it finds the key... if the upper bound is
-    # out of range, it will return an error).
-
-    _find_node($self, $self->{ROOT}, _split_bounds($self, $key));
+    # (it will use both bounds when searching the keys, and if the bounds
+    # are different it will check if the bounds are out of the node's
+    # range.)
 
+    my ($node, $parent) =
+      _find_node_parent($self, $self->{ROOT}, undef,
+			_split_bounds($self, $key));
+    return ($node) ? $node->{VALUE} : undef;
   }
 
 sub EXISTS                              # Check if a node exists
   {
     my ($self, $key) = @_;
-    defined( FETCH $self, $key );       # if it's fetchable...
+    my ($node, $parent) =
+      _find_node_parent($self, $self->{ROOT}, undef,
+			_split_bounds($self, $key));
+    return (defined($node));
   }
 
 sub STORE                               # Add a node
@@ -304,6 +353,62 @@ sub CLEAR                               # Wipe tree
     $self->{ROOT} = undef;
   }
 
+sub DELETE
+  {
+    my ($self, $key) = @_;
+
+    my ($lower_bound, $upper_bound) = _split_bounds($self, $key);
+    my ($node, $parent) =
+      _find_node_parent($self, $self->{ROOT}, undef,
+			$lower_bound, $upper_bound);
+
+    # Caveat: if $parent is not the parent of $node, you've got a problem!
+
+    unless ($node) { return; } # if node not found, nothing to delete
+
+    if ((&{$self->{COMPARE}}($lower_bound, $node->{KEY_LOW})) or
+	(&{$self->{COMPARE}}($upper_bound, $node->{KEY_HIGH})))
+      {
+	warnings::warn 
+	  ("Key range \`" .
+	   _join_bounds($self, $lower_bound, $upper_bound) .
+	   "\' is not a the defined key range \`" .
+	   _join_bounds($self, $node->{KEY_LOW}, $node->{KEY_HIGH}) .
+	   "\'" ),
+	     if (warnings::enabled);
+	return;	    
+      }
+
+    if ($parent)
+      {
+	if (&{$self->{COMPARE}}($parent->{KEY_HIGH}, $node->{KEY_LOW})<0)
+	  {
+
+	    $parent->{NODE_RIGHT} = $node->{NODE_RIGHT};
+
+	    $parent->{NODE_RIGHT} =
+	      _add_node($self, $parent->{NODE_RIGHT}, $node->{NODE_LEFT}),
+	        if ($node->{NODE_LEFT});
+	      }
+	elsif (&{$self->{COMPARE}}($parent->{KEY_LOW},
+				   $node->{KEY_HIGH})>0)
+	  {
+
+	    $parent->{NODE_LEFT} = $node->{NODE_LEFT};
+
+	    $parent->{NODE_LEFT} =
+	      _add_node($self, $parent->{NODE_LEFT}, $node->{NODE_RIGHT}),
+	        if ($node->{NODE_RIGHT});
+	  }
+      }
+    else
+      {	
+	$self->{ROOT} = $node->{NODE_LEFT};
+	$self->{ROOT} = _add_node($self, $self->{ROOT}, $node->{NODE_RIGHT}),
+	  if ($node->{NODE_RIGHT});
+      }
+    return $node->{VALUE};
+  }
 
 1;
 __END__
@@ -319,9 +424,10 @@ latter. It should work in Perl 5.005, although I have not tested it.
 
 It uses the following modules:
 
-  Carp                  # this may change in the future
+  Carp
   Tie::Hash
-  Time::HiRes           # for test.pl
+
+The test suite will use C<Time::HiRes> if it is available.
 
 =head2 Installation
 
@@ -403,12 +509,21 @@ or
    };
 
 Note that if you define it as a regular expression, warnings and errors
-will use the default comma ',' separator.
+will use the default comma ',' separator (since there is no way to "reverse"
+a regular expression).
 
 Duplicate and overlapping ranges are not supported. Once a range is defined,
 it exists for the lifetime of the hash. (Future versions may allow you to
 change this behavior.)
 
+Warnings are now disabled by default unless you run Perl with the -W flag.
+In theory, you should also be able to say
+
+  use warnings 'Tie::RangeHash';
+
+but this does not always seem to work. (Apparently something is broken
+with warnings.)
+
 Internally, the hash is actually a binary tree. Values are retrieved by
 searching the tree for nodes that where the key is within range.
 
@@ -422,7 +537,7 @@ array.) It appears to work, but has not been fully tested.
 A future version of this module may use an improved binary-tree algorithm.
 Or it may use something else.
 
-This module is incomplete... It needs the DELETE, FIRSTKEY, NEXTKEY,
+This module is incomplete... It needs the FIRSTKEY, NEXTKEY,
 and (maybe) DESTROY methods.
 
 =head1 AUTHOR
